/*-
 * main.c --
 *	The main file for this entire program. Exit routines etc
 *	reside here.
 *
 * Copyright (c) 1988 by the Regents of the University of California
 * Copyright (c) 1988 by Adam de Boor
 *
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appears in all copies.  Neither the University of California nor
 * Adam de Boor makes any representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 * Utility functions defined in this file:
 *	Main_ParseArgLine   	Takes a line of arguments, breaks them and
 *	    	  	    	treats them as if they were given when first
 *	    	  	    	invoked. Used by the parse module to implement
 *	    	  	    	the .MFLAGS target.
 *
 *	Error	  	    	Print a tagged error message. The global
 *	    	  	    	MAKE variable must have been defined. This
 *	    	  	    	takes a format string and two optional
 *	    	  	    	arguments for it.
 *
 *	Fatal	  	    	Print an error message and exit. Also takes
 *	    	  	    	a format string and two arguments.
 *
 *	Punt	  	    	Aborts all jobs and exits with a message. Also
 *	    	  	    	takes a format string and two arguments.
 *
 *	Finish	  	    	Finish things up by printing the number of
 *	    	  	    	errors which occured, as passed to it, and
 *	    	  	    	exiting.
 */
#ifndef lint
static char     *rcsid = "$Id: main.c,v 1.62 89/01/08 16:29:11 adam Exp $ SPRITE (Berkeley)";
#endif lint

#include    <stdio.h>
#ifdef Sprite
#include    <stdlib.h>
#endif /* Sprite */
#include    <sys/types.h>
#include    <sys/signal.h>
#include    <sys/stat.h>
#include    <fcntl.h>
#include    <sys/errno.h>
#include    <ctype.h>
#include    "make.h"
#include    <option.h>
extern int errno;


#ifndef DEFMAXLOCAL
#define DEFMAXLOCAL DEFMAXJOBS
#endif  DEFMAXLOCAL

#define MAKEFLAGS  	".MAKEFLAGS"

static char 	  	*progName;  	/* Our invocation name */
static Boolean	  	lockSet;    	/* TRUE if we set the lock file */
Lst			create;	    	/* Targets to be made */
time_t			now;	    	/* Time at start of make */
GNode			*DEFAULT;   	/* .DEFAULT node */
Boolean	    	    	allPrecious;	/* .PRECIOUS given on line by itself */

static int              printGraph;	/* -p flag */
static Boolean          noBuiltins;	/* -r flag */
static Boolean	  	noLocking;      /* -l flag */
static char		*makefile;	/* The makefile to read */
static int		maxJobs;	/* -J argument */
static int  	  	maxLocal;  	/* -L argument */
Boolean	    	  	debug;	    	/* -d flag */
Boolean	  	  	amMake; 	/* -M flag */
Boolean	    	  	noWarnings; 	/* -W flag */
Boolean	    	    	noExecute;  	/* -n flag */
Boolean	    	    	keepgoing;  	/* -k flag */
Boolean			queryFlag;  	/* -q flag */
Boolean			touchFlag;  	/* -t flag */
Boolean			usePipes;   	/* !-P flag */
Boolean			backwards;  	/* -B flag */
Boolean			ignoreErrors;	/* -i flag */
Boolean			beSilent;   	/* -s flag */
Boolean	    	    	sysVmake;   	/* -v flag */
Boolean			oldVars;    	/* -V flag */
static Boolean	  	XFlag=FALSE;   	/* -X flag given */
static Boolean	  	xFlag=FALSE;	/* -x flag given */
Boolean	    	  	noExport;   	/* Set TRUE if shouldn't export */
static Boolean	  	jobsRunning;	/* TRUE if the jobs might be running */

static int	  	MainDefineVar(),
			MainAddInclude(),
			MainDebug(),
			MainSysVInit(),
			MainResetCompat();
static Boolean	    	ReadMakefile();


static Option	  args[] = {
{	OPT_TRUE, 	"B",	(Address)&backwards,
 	"Be as backwards-compatible with make as possible without being make"},
{	OPT_FUNC, 	"C",	(Address)MainResetCompat,
 	"Cancel any current indications of compatibility and be PMake"},
{	OPT_FUNC, 	"D",	(Address)MainDefineVar,
 	"Define a variable with value \"1\""},
{	OPT_FUNC, 	"I",	(Address)MainAddInclude,
	"Specify (another) directory to search for included makefiles"},
{	OPT_INT,  	"J",	(Address)&maxJobs,
 	"Specify maximum overall concurrency"},
{	OPT_INT,  	"L",	(Address)&maxLocal,
 	"Specify maximum local concurrency"},
{	OPT_TRUE, 	"M",	(Address)&amMake,
 	"Be Make"},
{	OPT_FALSE,	"P",	(Address)&usePipes,
 	"Don't use pipes to catch the output of jobs, use files"},
{	OPT_TRUE, 	"V",	(Address)&oldVars,
 	"Use old-style variable substitution"},
{	OPT_TRUE, 	"W",	(Address)&noWarnings,
 	"Don't print warning messages when reading the makefile"},
{	OPT_TRUE, 	"X",	(Address)&XFlag,
	"Don't export any jobs"},
{	OPT_FUNC, 	"d",	(Address)MainDebug,
 	"Turn on debugging output"},
{	OPT_STRING,	"f",	(Address)&makefile,
	 "Specify a makefile to use instead of the default ones"},
{	OPT_TRUE, 	"i",	(Address)&ignoreErrors,
 	"Ignore any errors from executed commands"},
{	OPT_TRUE, 	"k",	(Address)&keepgoing,
 	"Keep working on unaffected targets on an error"},
#ifdef DONT_LOCK
{	OPT_FALSE, 	"l",	(Address)&noLocking,
 	"Turn on locking of the current directory."},
#else
{	OPT_TRUE, 	"l",	(Address)&noLocking,
 	"Turn off locking of the current directory."},
#endif /* DONT_LOCK */
{	OPT_TRUE, 	"n",	(Address)&noExecute,
 	"Don't execute commands to create targets, just print them"},
{	OPT_INT,  	"p",	(Address)&printGraph,
 	"Tell when to print the input graph: 1 (before), 2 (after) or 3 (both)."},
{	OPT_TRUE, 	"q",	(Address)&queryFlag,
 	"Just see if something would be done and exit non-zero if so"},
{	OPT_TRUE, 	"r",	(Address)&noBuiltins,
 	"Do not read the system makefile for pre-defined rules"},
{	OPT_TRUE, 	"s",	(Address)&beSilent,
 	"Don't print anything except error messages when creating targets"},
{	OPT_TRUE, 	"t",	(Address)&touchFlag,
 	"Only touch out-of-date targets, don't try and make them"},
{	OPT_FUNC,    	"v",	(Address)MainSysVInit,
	"Be compatible with System V make. Implies -B, -V and no locking"},
{	OPT_TRUE, 	"x",	(Address)&xFlag,
	"Allow exportation." },
{	OPT_DOC,  	"h",	(Address)NULL,
 	"System Configuration Information:"},
{	OPT_DOC,  	"",	(Address)NULL,
 	SYSPATHDOC},
{	OPT_DOC,  	"",	(Address)NULL,
	SHELLDOC},
{	OPT_DOC,  	"",	(Address)NULL,
	SYSMKDOC},
};

/*-
 *----------------------------------------------------------------------
 * MainParseArgs --
 *	Parse a given argument vector. Called from main() and from
 *	Main_ParseArgLine() when the .MAKEFLAGS target is used.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Various global and local flags will be set depending on the flags
 *	given
 *----------------------------------------------------------------------
 */
static void
MainParseArgs (argc, argv)
    int		  argc;	      /* Number of arguments in argv */
    char	  **argv;     /* The arguments themselves */
{
    register int  i;
    register char *cp;

    argc = Opt_Parse (argc, argv, args, Opt_Number(args),
		      OPT_ALLOW_CLUSTERING);

    /*
     * Take care of encompassing compatibility levels...
     */
    if (amMake) {
	backwards = TRUE;
    }
    if (backwards) {
	oldVars = TRUE;
    }

    /*
     * See if the rest of the arguments are variable assignments and perform
     * them if so. Else take them to be targets and stuff them on the end
     * of the "create" list.
     */
    for (i = 1; i < argc; i++) {
	if (Parse_IsVar (argv[i])) {
	    Parse_DoVar(argv[i], VAR_CMD);
	} else {
	    if (argv[i][0] == 0) {
		Punt("Bogus argument in MainParseArgs");
	    }
	    (void)Lst_AtEnd (create, (ClientData)argv[i]);
	}
    }
}

/*-
 *-----------------------------------------------------------------------
 * MainAddInclude --
 *	Note a new directory for included makefiles.
 *
 * Results:
 *	1, if a directory was specified.
 *
 * Side Effects:
 *	The -I flag is added to the MAKEFLAGS variable and the directory
 *	is given to the parse module.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
MainAddInclude (optKey, dir)
    char    	  *optKey;  	/* Option key that invoked this function */
    char    	  *dir;	    	/* Directory to add */
{
    char    	  *Iarg;

    if (dir == (char *)NULL) {
	fprintf(stderr, "Warning: -I needs an argument\n");
	return(0);
    } else {
	Parse_AddIncludeDir (dir);
    
	Iarg = (char *)malloc (strlen (dir) + 4);
	sprintf (Iarg, "-I %s", dir);
	Var_Append (MAKEFLAGS, Iarg, VAR_GLOBAL);
	free (Iarg);
	return(0);
    }
}

/*-
 *-----------------------------------------------------------------------
 * MainDefineVar --
 *	Define a variable in the global space with an empty value.
 *
 * Results:
 *	1, if a variable was specified.
 *
 * Side Effects:
 *	The variable is defined.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
MainDefineVar (optKey, var)
    char    	  *optKey;
    char    	  *var;
{
    char    	  *Darg;

    if (var == (char *)NULL) {
	fprintf(stderr, "Warning: -D needs an argument\n");
	return(0);
    } else {
	Var_Set (var, "1", VAR_GLOBAL);
	
	Darg = (char *)malloc (strlen (var) + 4);
	sprintf (Darg, "-D %s", var);
	Var_Append (MAKEFLAGS, Darg, VAR_GLOBAL);
	free (Darg);
	return(1);
    }
}

/*-
 *-----------------------------------------------------------------------
 * MainDebug --
 *	Turn on debugging for one or more modules.
 *
 * Results:
 *	1 if a module string was given.
 *
 * Side Effects:
 *	bits in 'debug' are altered.
 *
 *-----------------------------------------------------------------------
 */
static int
MainDebug(key, modules)
    char    	  *key;
    char    	  *modules;
{
    if ((modules != (char *)NULL) && (*modules != '-')) {
	while (*modules) {
	    switch (*modules) {
		case 's':
		    debug |= DEBUG_SUFF;
		    break;
		case 'm':
		    debug |= DEBUG_MAKE;
		    break;
		case 'j':
		    debug |= DEBUG_JOB;
		    break;
		case 't':
		    debug |= DEBUG_TARG;
		    break;
		case 'd':
		    debug |= DEBUG_DIR;
		    break;
		case 'v':
		    debug |= DEBUG_VAR;
		    break;
		case 'c':
		    debug |= DEBUG_COND;
		    break;
		case 'p':
		    debug |= DEBUG_PARSE;
		    break;
		case 'r':
		    debug |= DEBUG_RMT;
		    break;
		case 'a':
		    debug |= DEBUG_ARCH;
		    break;
	    }
	    modules++;
	}
	return(1);
    } else {
	debug = ~0;
	return(0);
    }
}

/*-
 *-----------------------------------------------------------------------
 * MainSysVInit --
 *	Initialize compatibility with System V Make.
 *
 * Results:
 *	0
 *
 * Side Effects:
 *	The following flags are set:
 *	    oldVars, backwards, sysVmake, noLocking.
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
MainSysVInit(optKey, arg)
    char    	*optKey;
    char    	*arg;
{
    oldVars = TRUE;
    backwards = TRUE;
    sysVmake = TRUE;
    noLocking = TRUE;
    return(0);
}

/*-
 *-----------------------------------------------------------------------
 * MainResetCompat --
 *	Reset all compatibility modes. This is used to control the
 *	exact amount of compatibility required by a makefile.
 *
 * Results:
 *	0
 *
 * Side Effects:
 *	amMake, sysVmake, backwards and oldVars are all set FALSE
 *
 *-----------------------------------------------------------------------
 */
/*ARGSUSED*/
static int
MainResetCompat(optKey, arg)
    char    	*optKey;
    char    	*arg;
{
    oldVars = backwards = sysVmake = amMake = FALSE;

    return(0);
}

/*-
 *----------------------------------------------------------------------
 * Main_ParseArgLine --
 *  	Used by the parse module when a .MFLAGS or .MAKEFLAGS target
 *	is encountered and by main() when reading the .MAKEFLAGS envariable.
 *	Takes a line of arguments and breaks it into its
 * 	component words and passes those words and the number of them to the
 *	MainParseArgs function.
 *	The line should have all its leading whitespace removed.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	Only those that come from the various arguments.
 *-----------------------------------------------------------------------
 */
void
Main_ParseArgLine (line)
    char    	  *line;      /* Line to fracture */
{
    char    	  **argv;     /* Manufactured argument vector */
    int     	  argc;	      /* Number of arguments in argv */

    if (line == NULL) return;
    while (*line == ' ') line++;

    argv = Str_BreakString (line, " \t", "\n", &argc);

    MainParseArgs(argc, argv);
    /*
     * XXX: Should call Str_FreeVec but Opt_Parse overwrites the vector.
     * I could copy the vector, pass that to MainParseArgs and be able to
     * free both of them, but that strikes me as silly...
     */
    free((Address)argv);
}

/*-
 *-----------------------------------------------------------------------
 * MainUnlock --
 *	Unlock the current directory. Called as an ExitHandler.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The locking file LOCKFILE is removed.
 *
 *-----------------------------------------------------------------------
 */
static void
MainUnlock ()
{
    (void)unlink (LOCKFILE);
}

/*-
 *----------------------------------------------------------------------
 * main --
 *	The main function, for obvious reasons. Initializes variables
 *	and a few modules, then parses the arguments give it in the
 *	environment and on the command line. Reads the system makefile
 *	followed by either Makefile, makefile or the file given by the
 *	-f argument. Sets the .MAKEFLAGS PMake variable based on all the
 *	flags it has received by then uses either the Make or the Compat
 *	module to create the initial list of targets.
 *
 * Results:
 *	If -q was given, exits -1 if anything was out-of-date. Else it exits
 *	0.
 *
 * Side Effects:
 *	The program exits when done. Targets are created. etc. etc. etc.
 *
 *----------------------------------------------------------------------
 */
main (argc, argv)
    int           argc;
    char          **argv;
{
    Lst           targs;      	/* list of target nodes to create. Passed to
				 * Make_Init */
    Boolean       outOfDate;  	/* FALSE if all targets up to date */
    char    	  *cp;


    create = Lst_Init (FALSE);

    beSilent = FALSE;	      	/* Print commands as executed */
    ignoreErrors = FALSE;     	/* Pay attention to non-zero returns */
    noExecute = FALSE;	      	/* Execute all commands */
    keepgoing = FALSE;	      	/* Stop on error */
    allPrecious = FALSE;      	/* Remove targets when interrupted */
    queryFlag = FALSE;	      	/* This is not just a check-run */
    noBuiltins = FALSE;	      	/* Read the built-in rules */
    touchFlag = FALSE;	      	/* Actually update targets */
    usePipes = TRUE;	      	/* Catch child output in pipes */
#ifndef DONT_LOCK
    noLocking = FALSE;	      	/* Lock the current directory against other
				 * pmakes */
#else
    noLocking = TRUE;
#endif /* DONT_LOCK */
    debug = 0;	      	    	/* No debug verbosity, please. */
    noWarnings = FALSE;	    	/* Print warning messages */
    sysVmake = FALSE;	    	/* Don't be System V compatible */

    jobsRunning = FALSE;

    makefile = (char *) NULL; 	/* Use default makefiles */
    maxJobs = DEFMAXJOBS;     	/* Set the default maximum concurrency */
    maxLocal = DEFMAXLOCAL;   	/* Set the default local max concurrency */
    
    /*
     * See what the user calls us. If s/he calls us (yuck) "make", then
     * act like it. Otherwise act like our normal, cheerful self.
     */
    cp = rindex (argv[0], '/');
    if (cp != (char *)NULL) {
	cp += 1;
    } else {
	cp = argv[0];
    }
    progName = cp;

    if (strcmp (cp, "make") == 0) {
	amMake = TRUE;	      	/* Be like make */
	backwards = TRUE;     	/* Do things the old-fashioned way */
	oldVars = TRUE;	      	/* Same with variables */
    } else if (strcmp(cp, "smake") == 0 || strcmp(cp, "vmake") == 0) {
	MainSysVInit("", "");
    } else {
	amMake = FALSE;
	backwards = FALSE;    	/* Do things MY way, not MAKE's */
	oldVars = FALSE;      	/* don't substitute for undefined variables */
    }

    /*
     * Initialize the parsing, directory and variable modules to prepare
     * for the reading of inclusion paths and variable settings on the
     * command line 
     */
    Dir_Init ();		/* Initialize directory structures so -I flags
				 * can be processed correctly */
    Parse_Init ();		/* Need to initialize the paths of #include
				 * directories */
    Var_Init ();		/* As well as the lists of variables for
				 * parsing arguments */

    /*
     * Initialize various variables.
     *	.PMAKE gets how we were executed.
     *	MAKE also gets this name, for compatibility
     *	.MAKEFLAGS gets set to the empty string just in case.
     *  MFLAGS also gets initialized empty, for compatibility.
     */
    Var_Set (".PMAKE", argv[0], VAR_GLOBAL);
    Var_Set ("MAKE", argv[0], VAR_GLOBAL);
    Var_Set (MAKEFLAGS, "", VAR_GLOBAL);
    Var_Set ("MFLAGS", "", VAR_GLOBAL);

    /*
     * First snag any flags out of the PMAKE environment variable.
     * (Note this is *not* MAKEFLAGS since /bin/make uses that and it's in
     * a different format).
     */

    Main_ParseArgLine (getenv("PMAKE"));
    
    MainParseArgs (argc, argv);

    /*
     * XXX: Should be able to specify more than one default system path by
     *	    using standard 'directory1:...:directoryn' format, I think
     */
    Parse_AddIncludeDir (DEFSYSPATH);

    /*
     * If the user didn't tell us not to lock the directory, attempt to create
     * the lock file. Complain if we can't, otherwise set up an exit handler
     * to remove the lock file...
     */
    if (!noLocking) {
	int	  	oldMask;    /* Previous signal mask */
	int	  	lockID;     /* Stream ID of opened lock file */
	
#ifndef SYSV
	oldMask = sigblock(sigmask(SIGINT));
#else
	oldMask = sighold(SIGINT);
#endif
	
	lockID = open (LOCKFILE, O_CREAT | O_EXCL, 0666);
	if (lockID < 0 && errno == EEXIST) {
	    /*
	     * Find out who owns the file. If the user who called us
	     * owns it, then we ignore the lock file. Note that we also
	     * do not install an exit handler to remove the file -- if the
	     * lockfile is there from a previous make, it'll still be there
	     * when we leave.
	     */
	    struct stat   fsa;    /* Attributes of the lock file */

	    (void) stat (LOCKFILE,  &fsa);
	    if (fsa.st_uid == getuid()) {
		Error ("Lockfile owned by you -- ignoring it");
		lockSet = FALSE;
	    } else {
		char  	lsCmd[40];
		(void)sprintf (lsCmd, "ls -l %s", LOCKFILE);
		(void)system(lsCmd);
		Fatal ("This directory is already locked (%s exists)",
		       LOCKFILE);
	    }
	} else if (lockID < 0) {
	    Fatal ("Could not create lock file %s", LOCKFILE);
	} else {
	    extern exit();

	    lockSet = TRUE;
#ifdef sun
	    on_exit(MainUnlock);
#endif
	    /* (void) Proc_SetExitHandler (MainUnlock, (ClientData)0); */
	    signal(SIGINT, exit);
	    (void)close (lockID);
	}
	
#ifndef SYSV
	(void) sigsetmask(oldMask);
#else 
	(void) sigrelse(SIGINT);
#endif
    }

    /*
     * Initialize archive, target and suffix modules in preparation for
     * parsing the makefile(s) 
     */
    Arch_Init ();
    Targ_Init ();
    Suff_Init ();

    DEFAULT = NILGNODE;

    now = time(0);

    /*
     * Read in the built-in rules first, followed by the specified makefile,
     * if it was (makefile != (char *) NULL), or the default Makefile and
     * makefile, in that order, if it wasn't. 
     */
    if (!noBuiltins && !ReadMakefile (DEFSYSMK)) {
	Fatal ("Could not open system rules (%s)", DEFSYSMK);
    }

    if ((makefile != (char *) NULL) && !ReadMakefile (makefile)) {
	Fatal ("Cannot open %s", makefile);
    } else if ((makefile == (char *) NULL) &&
	       !ReadMakefile ((amMake || sysVmake) ? "makefile" : "Makefile"))
    {
	(void) ReadMakefile ((amMake  || sysVmake) ? "Makefile" : "makefile");
    }

    /*
     * Setup the MAKEFLAGS variable from the way our flag variables are
     * initialized. -I and -D flags have already been done.
     */

    if (amMake) {
	noExport = !xFlag || XFlag;
	Var_Append (MAKEFLAGS, "-M", VAR_GLOBAL);
    } else {
	noExport = XFlag && !xFlag;
    }
    
    if (backwards && !amMake) {
	Var_Append (MAKEFLAGS, "-B", VAR_GLOBAL);
    }
    if (maxJobs != DEFMAXJOBS) {
	char Jarg[10];

	sprintf (Jarg, "-J %d", maxJobs);
	Var_Append (MAKEFLAGS, Jarg, VAR_GLOBAL);
    }
    if (maxLocal != DEFMAXLOCAL) {
	char Larg[10];

	sprintf (Larg, "-L %d", maxLocal);
	Var_Append (MAKEFLAGS, Larg, VAR_GLOBAL);
    }
    if (!usePipes) {
	Var_Append (MAKEFLAGS, "-P", VAR_GLOBAL);
    }
    if (sysVmake) {
	Var_Append(MAKEFLAGS, "-v", VAR_GLOBAL);
    }
    if (oldVars && !amMake && !backwards) {
	Var_Append (MAKEFLAGS, "-V", VAR_GLOBAL);
    }
    if (noWarnings) {
	Var_Append (MAKEFLAGS, "-W", VAR_GLOBAL);
    }
    if (XFlag) {
	Var_Append (MAKEFLAGS, "-X", VAR_GLOBAL);
    }
    if (xFlag) {
	Var_Append (MAKEFLAGS, "-x", VAR_GLOBAL);
    }
    if (debug) {
	char	  modules[20];

	cp = modules;
	if (DEBUG(SUFF)) {
	    *cp++ = 's';
	}
	if (DEBUG(MAKE)) {
	    *cp++ = 'm';
	}
	if (DEBUG(JOB)) {
	    *cp++ = 'j';
	}
	if (DEBUG(TARG)) {
	    *cp++ = 't';
	}
	if (DEBUG(DIR)) {
	    *cp++ = 'd';
	}
	if (DEBUG(VAR)) {
	    *cp++ = 'v';
	}
	if (DEBUG(COND)) {
	    *cp++ = 'c';
	}
	if (DEBUG(PARSE)) {
	    *cp++ = 'p';
	}
	if (DEBUG(RMT)) {
	    *cp++ = 'r';
	}
	if (DEBUG(ARCH)) {
	    *cp++ = 'a';
	}
	*cp = '\0';
	Var_Append (MAKEFLAGS,
		    Str_Concat("-d", modules, STR_ADDSPACE),
		    VAR_GLOBAL);
    }
    if (ignoreErrors) {
	Var_Append (MAKEFLAGS, "-i", VAR_GLOBAL);
    }
    if (keepgoing) {
	Var_Append (MAKEFLAGS, "-k", VAR_GLOBAL);
    }
    if (noLocking) {
	Var_Append (MAKEFLAGS, "-l", VAR_GLOBAL);
    }
    if (noExecute) {
	Var_Append (MAKEFLAGS, "-n", VAR_GLOBAL);
    }
    if (printGraph) {
	char parg[10];

	sprintf (parg, "-p %d", printGraph);
	Var_Append (MAKEFLAGS, parg, VAR_GLOBAL);
    }
    if (noBuiltins) {
	Var_Append (MAKEFLAGS, "-r", VAR_GLOBAL);
    }
    if (beSilent) {
	Var_Append (MAKEFLAGS, "-s", VAR_GLOBAL);
    }
    if (touchFlag) {
	Var_Append (MAKEFLAGS, "-t", VAR_GLOBAL);
    }
    Var_Append ("MFLAGS", Var_Value(MAKEFLAGS, VAR_GLOBAL), VAR_GLOBAL);

    /*
     * Install all the flags into the PMAKE envariable.
     */
    setenv("PMAKE", Var_Value(MAKEFLAGS, VAR_GLOBAL));

    /*
     * For compatibility, look at the directories in the VPATH variable
     * and add them to the search path, if the variable is defined. The
     * variable's value is in the same format as the PATH envariable, i.e.
     * <directory>:<directory>:<directory>...
     */
    if (Var_Exists ("VPATH", VAR_CMD)) {
	char	  *vpath;
	char	  *path;
	char  	  *cp;
	char  	  savec;
	static char VPATH[] = "${VPATH}";   /* GCC stores string constants in
					     * read-only memory, but Var_Subst
					     * will want to write this thing,
					     * so store it in an array */
	
	vpath = Var_Subst (VPATH, VAR_CMD, FALSE);

	path = vpath;
	do {
	    /*
	     * Skip to end of directory
	     */
	    for (cp = path; *cp != ':' && *cp != '\0'; cp++) {
		continue;
	    }
	    /*
	     * Save terminator character to figure out when to stop
	     */
	    savec = *cp;
	    *cp = '\0';
	    /*
	     * Add directory to search path
	     */
	    Dir_AddDir (path, dirSearchPath);
	    *cp = savec;
	    path = cp + 1;
	} while (savec == ':');
	free((Address)vpath);
    }
	    
    /*
     * Now that all search paths have been read for suffixes et al, it's
     * time to add the default search path to their lists...
     */
    Suff_DoPaths();

    /*
     * Print the initial graph, if the user requested it
     */
    if (printGraph & 1) {
	Targ_PrintGraph (1);
    }

    Rmt_Init();

    /*
     * Have now read the entire graph and need to make a list of targets to
     * create. If none was given on the command line, we consult the parsing
     * module to find the main target(s) to create.
     */
    if (Lst_IsEmpty (create)) {
	targs = Parse_MainName ();
    } else {
	targs = Targ_FindList (create, TARG_CREATE);
    }

    if (!amMake) {
	/*
	 * Initialize job module before traversing the graph, now that any
	 * .BEGIN and .END targets have been read. This is done only if the
	 * -q flag wasn't given (to prevent the .BEGIN from being executed
	 * should it exist).
	 */
	if (!queryFlag) {
	    if (maxLocal == -1) {
		maxLocal = maxJobs;
	    }
	    Job_Init (maxJobs, maxLocal);
	    jobsRunning = TRUE;
	}
	
	/*
	 * Traverse the graph, checking on all the targets 
	 */
	outOfDate = Make_Run (targs);
    } else {
	/*
	 * Compat_Init will take care of creating all the targets as well
	 * as initializing the module.
	 */
	Compat_Run(targs);
    }
    
    /*
     * Print the graph now it's been processed if the user requested it
     */
    if (printGraph & 2) {
	Targ_PrintGraph (2);
    }

    if (queryFlag && outOfDate) {
	exit (-1);
    } else {
	exit (0);
    }
}

/*-
 *-----------------------------------------------------------------------
 * ReadMakefile  --
 *	Open and parse the given makefile.
 *
 * Results:
 *	TRUE if ok. FALSE if couldn't open file.
 *
 * Side Effects:
 *	lots
 *-----------------------------------------------------------------------
 */
static Boolean
ReadMakefile (fname)
    char          *fname;     /* makefile to read */
{
    if (strcmp (fname, "-") == 0) {
	Parse_File ("(stdin)", stdin);
	return (TRUE);
    } else {
	FILE *	  stream;
	
	stream = fopen (fname, "r");
    
	if (stream == (FILE *) NULL) {
	    return (FALSE);
	}
	Parse_File (fname, stream);
	fclose (stream);
	return (TRUE);
    }
}

/*-
 *-----------------------------------------------------------------------
 * Error --
 *	Print an error message given its format and 0, 1, 2 or 3 arguments.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	The message is printed.
 *
 *-----------------------------------------------------------------------
 */
/*VARARGS1*/
void
Error (fmt, arg1, arg2, arg3)
    char    	  *fmt;	    	    /* Format string */
    int	    	  arg1,	    	    /* First optional argument */
		  arg2,	    	    /* Second optional argument */
		  arg3;	    	    /* Third optional argument */
{
    static char   estr[BSIZE];	    /* output string */

    sprintf (estr, "%s: ", Var_Value(".PMAKE", VAR_GLOBAL));
    sprintf (&estr[strlen (estr)], fmt, arg1, arg2, arg3);
    (void) strcat (estr, "\n");

    fputs (estr, stderr);
    fflush (stderr);
}

/*-
 *-----------------------------------------------------------------------
 * Fatal --
 *	Produce a Fatal error message. If jobs are running, waits for them
 *	to finish.
 *
 * Results:
 *	None
 *
 * Side Effects:
 *	The program exits
 *-----------------------------------------------------------------------
 */
/* VARARGS1 */
void
Fatal (fmt, arg1, arg2)
    char          *fmt;	      	  /* format string */
    int           arg1;	      	  /* first optional argument */
    int           arg2;	      	  /* second optional argument */
{
    if (jobsRunning) {
	Job_Wait();
    }
    
    Error (fmt, arg1, arg2);

    if (printGraph & 2) {
	Targ_PrintGraph(2);
    }
    exit (1);
}

/*
 *-----------------------------------------------------------------------
 * Punt --
 *	Major exception once jobs are being created. Kills all jobs, prints
 *	a message and exits.
 *
 * Results:
 *	None 
 *
 * Side Effects:
 *	All children are killed indiscriminately and the program Lib_Exits
 *-----------------------------------------------------------------------
 */
/* VARARGS1 */
void
Punt (fmt, arg1, arg2)
    char          *fmt;	/* format string */
    int           arg1;	/* optional argument */
    int	    	  arg2;	/* optional second argument */
{
    if (jobsRunning) {
	Job_AbortAll ();
    }
    Error (fmt, arg1, arg2);

    if (printGraph & 2) {
	Targ_PrintGraph(2);
    }
    
    exit (1);
}

/*
 *-----------------------------------------------------------------------
 * Finish --
 *	Called when aborting due to errors in child shell to signal
 *	abnormal exit. 
 *
 * Results:
 *	None 
 *
 * Side Effects:
 *	The program exits
 * -----------------------------------------------------------------------
 */
void
Finish (errors)
    int             errors;	/* number of errors encountered in Make_Make */
{
    Fatal ("%d error%s", errors, errors == 1 ? "" : "s");
}

#if defined(SYSV) || defined(is68k) && !defined(PROF)

exit(status)
{
    if (lockSet) {
	MainUnlock();
    }
    _cleanup();
    _exit(status);
}
#endif /* System V */
